#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from enum import Enum, auto

from th2_common.schema.configuration.abstract_configuration import AbstractConfiguration

from abc import ABC, abstractmethod
from typing import List


class FieldFilterOperation(Enum):
    EQUAL = auto()
    NOT_EQUAL = auto()
    EMPTY = auto()
    NOT_EMPTY = auto()
    WILDCARD = auto()


class FieldFilterConfiguration(AbstractConfiguration):

    def __init__(self, value: str = None, expectedValue: str = None, fieldName: str = None,
                 operation: FieldFilterOperation = None, **kwargs) -> None:
        self.value = value or expectedValue
        self.field_name = fieldName
        self.operation = FieldFilterOperation[operation]
        self.check_unexpected_args(kwargs)

    def __str__(self):
        return f'Value: {self.value} Field name: {self.field_name} Operation: {self.operation}'


class RouterFilterConfiguration(AbstractConfiguration, ABC):

    @abstractmethod
    def get_metadata(self) -> List[FieldFilterConfiguration]:
        pass

    @abstractmethod
    def get_message(self) -> List[FieldFilterConfiguration]:
        pass


class MqRouterFilterConfiguration(RouterFilterConfiguration):

    def __init__(self, metadata=None, message=None, **kwargs) -> None:

        self.metadata = []
        self.message = []

        if isinstance(metadata, dict) and isinstance(message, dict):
            self.metadata = [FieldFilterConfiguration(**metadata[key], fieldName=key) for key in metadata.keys()]
            self.message = [FieldFilterConfiguration(**message[key], fieldName=key) for key in message.keys()]

        elif isinstance(metadata, list) and isinstance(message, list):
            self.metadata = [FieldFilterConfiguration(**key) for key in metadata]
            self.message = [FieldFilterConfiguration(**key) for key in message]

        self.check_unexpected_args(kwargs)

    def get_metadata(self) -> List[FieldFilterConfiguration]:
        return self.metadata

    def get_message(self) -> List[FieldFilterConfiguration]:
        return self.message

    def __str__(self):
        return self.metadata + self.message


class QueueConfiguration(AbstractConfiguration):

    def __init__(self, name: str, queue: str, exchange: str, attributes: list, filters: list, can_read=True,
                 can_write=True, **kwargs) -> None:
        self.routing_key = name
        self.queue = queue
        self.exchange = exchange
        self.attributes = attributes
        self.filters = [MqRouterFilterConfiguration(**filter_schema) for filter_schema in filters]
        self.can_read = can_read
        self.can_write = can_write
        self.check_unexpected_args(kwargs)


class MessageRouterConfiguration(AbstractConfiguration):
    def __init__(self, queues: dict, **kwargs) -> None:
        self.queues = {queue_alias: QueueConfiguration(**queues[queue_alias]) for queue_alias in queues.keys()}
        self.check_unexpected_args(kwargs)

    def get_queue_by_alias(self, queue_alias):
        return self.queues[queue_alias]

    def find_queues_by_attr(self, attrs) -> {str: QueueConfiguration}:
        result = dict()
        for queue_alias in self.queues.keys():
            if all(attr in self.queues[queue_alias].attributes for attr in attrs):
                result[queue_alias] = self.queues[queue_alias]
        return result


class ConnectionManagerConfiguration(AbstractConfiguration):
    def __init__(self, subscriberName=None,
                 connectionTimeout=-1,
                 connectionCloseTimeout = 10000,
                 maxRecoveryAttempts=5,
                 minConnectionRecoveryTimeout=10000,
                 maxConnectionRecoveryTimeout=60000,
                 prefetchCount=10,
                 messageRecursionLimit=100,
                 **kwargs):
        self.subscriber_name = subscriberName
        self.connection_timeout = int(connectionTimeout)
        self.connection_close_timeout = int(connectionCloseTimeout)
        self.max_recovery_attempts = int(maxRecoveryAttempts)
        self.min_connection_recovery_timeout = int(minConnectionRecoveryTimeout)
        self.max_connection_recovery_timeout = int(maxConnectionRecoveryTimeout)
        self.prefetch_count = int(prefetchCount)
        self.message_recursion_limit = int(messageRecursionLimit)
        self.check_unexpected_args(kwargs)



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import asyncio
import functools
import logging
import queue
import threading
import time
from typing import Optional

import pika
from pika.adapters.asyncio_connection import AsyncioConnection
from pika.channel import Channel

logger = logging.getLogger(__name__)


class ReconnectingPublisher(object):
    TIMEOUT_WAIT_CONFIRM_DELIVERY_BY_CLOSE = 10
    TIMEOUT_STOPPING = 5

    def __init__(self, connection_parameters: pika.ConnectionParameters):
        self._connection_parameters = connection_parameters

        self._connection: Optional[AsyncioConnection] = None
        self._channel: Optional[Channel] = None
        self._loop = asyncio.new_event_loop()
        self._not_executed_tasks = queue.Queue(maxsize=1000)

        self._deliveries = []
        self._acked = None
        self._nacked = None
        self._message_number = None

        self._message_lock = threading.Lock()
        self._publishing_allowed = threading.Event()

        self._stopping = False

    def on_connection_open(self, _unused_connection):
        logger.info('Publisher\'s connection has been opened')
        self.open_channel()

    def on_connection_open_error(self, _unused_connection, err):
        self._publishing_allowed.clear()
        logger.error('An error occurred while opening a publisher\'s connection,'
                     ' reopening as soon as possible: %s', err)
        if not self._stopping and not self._connection.is_closed and not self._connection.is_closing:
            self._connection.close()

    def on_connection_closed(self, _unused_connection, reason):
        self._publishing_allowed.clear()
        logger.info('Publisher\'s connection has been closed: %s', reason)
        self._loop.call_soon_threadsafe(self._connection.ioloop.stop)

    def open_channel(self):
        logger.info('Creating a channel for Publisher')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        logger.info('Publisher\'s channel has been opened')
        self._channel = channel
        self._channel.confirm_delivery(self.on_delivery_confirmation)
        self._channel.add_on_close_callback(self.on_channel_closed)
        self._publishing_allowed.set()

    def on_channel_closed(self, channel, reason):
        self._publishing_allowed.clear()
        logger.info('Publisher\'s channel %i has been closed: %s', channel, reason)
        if not self._stopping and not self._connection.is_closed and not self._connection.is_closing:
            self._connection.close()

    def on_delivery_confirmation(self, method_frame):
        with self._message_lock:
            confirmation_type = method_frame.method.NAME.split('.')[1].lower()
            logger.debug('Received %s for delivery tag: %i multi: %s', confirmation_type,
                         method_frame.method.delivery_tag, method_frame.method.multiple)
            cnt = 0
            if method_frame.method.multiple:
                idx = len(self._deliveries) - 1
                while self._deliveries[idx] != method_frame.method.delivery_tag:
                    idx -= 1
                while idx >= 0:
                    self._deliveries.pop(0)
                    cnt += 1
                    idx -= 1
            else:
                cnt += 1
                self._deliveries.remove(method_frame.method.delivery_tag)
            if confirmation_type == 'ack':
                self._acked += cnt
            elif confirmation_type == 'nack':
                self._nacked += cnt

            logger.debug(
                'Published %i messages, %i have yet to be confirmed, '
                '%i were acked and %i were nacked', self._message_number,
                len(self._deliveries), self._acked, self._nacked)

    def publish_message(self, exchange_name, routing_key, message):
        self._publishing_allowed.wait()
        cb = functools.partial(self._basic_publish, exchange_name, routing_key, message)
        while not self._not_executed_tasks.empty():
            self._connection.ioloop.call_soon_threadsafe(self._not_executed_tasks.get())
        while self._stopping:
            # It may be worth throwing an exception here that the connection is closed manually.
            time.sleep(1)
        self._connection.ioloop.call_soon_threadsafe(cb)

    def _basic_publish(self, exchange_name, routing_key, message):
        with self._message_lock:
            if not self._publishing_allowed.is_set():
                cb = functools.partial(self._basic_publish, exchange_name, routing_key, message)
                self._not_executed_tasks.put(cb)
            else:
                self._channel.basic_publish(exchange=exchange_name,
                                            routing_key=routing_key,
                                            body=message)

                self._message_number += 1
                self._deliveries.append(self._message_number)
                logger.debug('Published message # %i', self._message_number)

    def connect(self):
        self._connection = AsyncioConnection(
            self._connection_parameters,
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed,
            custom_ioloop=self._loop)

    def run(self):
        while not self._stopping:
            self._connection = None
            with self._message_lock:
                self._acked = 0
                self._nacked = 0
                self._message_number = 0

            try:
                self.connect()
                logger.info('Publisher\'s Event loop running')
                self._connection.ioloop.run_forever()
                logger.info('Publisher\'s Event loop has been stopped')
            except Exception:
                logger.exception("Error while running Publisher")

    def stop(self):
        logger.info('Publisher stopping')
        self._stopping = True
        logger.info(f'Waiting for the remaining messages to be published for'
                    f' {ReconnectingPublisher.TIMEOUT_STOPPING} sec.')
        # Pull from queue tasks
        time.sleep(ReconnectingPublisher.TIMEOUT_STOPPING)
        deliveries_size = len(self._deliveries)
        wait_counter = 0
        while deliveries_size > 0 and wait_counter <= ReconnectingPublisher.TIMEOUT_WAIT_CONFIRM_DELIVERY_BY_CLOSE:
            logger.info(f'Deliveries size != 0, size={deliveries_size}. Wait confirm delivery '
                        f'{ReconnectingPublisher.TIMEOUT_WAIT_CONFIRM_DELIVERY_BY_CLOSE - wait_counter} sec.')
            time.sleep(1)
            wait_counter += 1
            deliveries_size = len(self._deliveries)
        self.close_channel()
        self.close_connection()

    def close_channel(self):
        if self._channel is not None and not self._channel.is_closing and not self._channel.is_closed:
            logger.info('Closing Publisher\'s channel')
            self._channel.close()

    def close_connection(self):
        if self._connection is not None and not self._connection.is_closing and not self._connection.is_closed:
            logger.info('Closing Publisher\'s connection')
            self._connection.close()



#   Copyright 2021-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List, Tuple

from google.protobuf.duration_pb2 import Duration
from google.protobuf.pyext._message import RepeatedCompositeContainer
from th2_grpc_common.common_pb2 import MessageID, MessageGroupBatch, AnyMessage, EventBatch, Direction, Value, \
    ListValue, Message, MessageMetadata, ConnectionID, MetadataFilter, RootMessageFilter, MessageFilter, \
    ValueFilter, ListValueFilter, SimpleList, RootComparisonSettings, MessageGroup

import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.message.configuration.message_configuration import MessageRouterConfiguration


def get_filters(configuration: MessageRouterConfiguration, aliases: List[str]):
    return [configuration.queues[alias].filters for alias in aliases]


def get_sequence(group: MessageGroup):
    if group.messages[0].HasField('raw_message'):
        return group.messages[0].raw_message.metadata.id.sequence
    elif group.messages[0].HasField('message'):
        return group.messages[0].message.metadata.id.sequence


def get_session_alias_and_direction(message_id: MessageID) -> Tuple[str, str]:
    return message_id.connection_id.session_alias, Direction.Name(message_id.direction)


def get_session_alias_and_direction_group(any_message: AnyMessage) -> Tuple[str, str]:
    if any_message.HasField('message'):
        return get_session_alias_and_direction(any_message.message.metadata.id)
    elif any_message.HasField('raw_message'):
        return get_session_alias_and_direction(any_message.raw_message.metadata.id)
    else:
        return common_metrics.UNKNOWN_LABELS


def get_debug_string(class_name: str, ids: List[MessageID]) -> str:
    session_alias, direction = get_session_alias_and_direction(ids[0])
    sequences = ', '.join([str(i.sequence) for i in ids])
    return f'{class_name}: {session_alias = }, {direction = }, {sequences = }'.strip()


def get_debug_string_event(event_batch: EventBatch) -> str:
    return f'EventBatch: {event_batch.events[0].id}'.strip()


def get_debug_string_group(group_batch: MessageGroupBatch) -> str:
    messages = [message for group in group_batch.groups for message in group.messages]
    session_alias, direction = get_session_alias_and_direction_group(messages[0])
    sequences = []
    for message in messages:
        if message.HasField('message'):
            sequences.append(str(message.message.metadata.id.sequence))
        elif message.HasField('raw_message'):
            sequences.append(str(message.raw_message.metadata.id.sequence))
    sequences = ''.join(sequences)

    return f'MessageGroupBatch: ' \
           f'{session_alias = }, ' \
           f'{direction = }, ' \
           f'{sequences = }'.strip()


def convert_message_value(value):
    if isinstance(value, Value):
        return value
    elif isinstance(value, (str, int, float)):
        return Value(simple_value=str(value))
    elif isinstance(value, list):
        return Value(list_value=ListValue(values=[convert_message_value(x) for x in value]))
    elif isinstance(value, dict):
        return Value(message_value=Message(fields={key: convert_message_value(value[key]) for key in value}))


def create_message(fields: dict, session_alias=None, message_type=None):
    return Message(metadata=MessageMetadata(id=MessageID(connection_id=ConnectionID(session_alias=session_alias)),
                                            message_type=message_type),
                   fields={field: convert_message_value(fields[field]) for field in fields})


def convert_filter_value(value, message_type=None, direction=None, values=False, metadata=False):
    if isinstance(value, ValueFilter):
        return value
    elif isinstance(value, (str, int, float)) and values is True:
        return ValueFilter(simple_filter=str(value))
    elif isinstance(value, (str, int, float)) and metadata is True:
        return MetadataFilter.SimpleFilter(value=str(value))
    elif isinstance(value, list) and values is True:
        return ValueFilter(list_filter=ListValueFilter(
            values=[convert_filter_value(x, values=values, metadata=metadata) for x in value]))
    elif isinstance(value, list) and metadata is True:
        return MetadataFilter.SimpleFilter(simple_list=SimpleList(simple_values=value))
    elif isinstance(value, dict):
        return ValueFilter(
            message_filter=MessageFilter(messageType=message_type,
                                         fields={key: convert_filter_value(value[key],
                                                                           values=values,
                                                                           metadata=metadata)
                                                 for key in value},
                                         direction=direction))


def create_root_message_filter(message_type=None,
                               message_filter=None,
                               metadata_filter=None,
                               ignore_fields: List[str] = None,
                               check_repeating_group_order: bool = None,
                               time_precision: Duration = None,
                               decimal_precision: str = None):
    if message_filter is None:
        message_filter = {}
    if metadata_filter is None:
        metadata_filter = {}
    return RootMessageFilter(messageType=message_type,
                             message_filter=MessageFilter(fields={
                                 field: convert_filter_value(message_filter[field], values=True)
                                 for field in message_filter}),
                             metadata_filter=MetadataFilter(property_filters={
                                 value: convert_filter_value(metadata_filter[value], metadata=True)
                                 for value in metadata_filter}),
                             comparison_settings=RootComparisonSettings(
                                 ignore_fields=ignore_fields,
                                 check_repeating_group_order=check_repeating_group_order,
                                 time_precision=time_precision,
                                 decimal_precision=decimal_precision))


def convert_value_into_typed_field(field_value, typed_field_value):
    if field_value.WhichOneof('kind') == 'simple_value':
        return type(typed_field_value)(field_value.simple_value)
    elif field_value.WhichOneof('kind') == 'list_value':
        return [convert_value_into_typed_field(list_item, typed_field_value.add())
                for list_item in field_value.list_value.values]
    elif field_value.WhichOneof('kind') == 'message_value':
        fields_typed = {field: convert_value_into_typed_field(field_value.message_value.fields[field],
                                                              getattr(typed_field_value, field))
                        for field in field_value.message_value.fields}
        return type(typed_field_value)(**fields_typed)


def create_typed_message_from_message(message, message_type):
    response_fields = [field.name for field in message_type().DESCRIPTOR.fields]
    fields_typed = {field: convert_value_into_typed_field(message.fields[field], getattr(message_type(), field))
                    for field in response_fields}
    return message_type(**fields_typed)



#   Copyright 2021-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
from typing import Set

from th2_common.schema.metrics.metric import Metric
from th2_common.schema.metrics.metric_monitor import MetricMonitor


class AbstractMetric(Metric, ABC):

    __disabled_monitors: Set[MetricMonitor] = set()

    @property
    def enabled(self) -> bool:
        return not self.__disabled_monitors

    def create_monitor(self, name: str) -> MetricMonitor:
        return MetricMonitor(name, self)

    def is_enabled(self, monitor: MetricMonitor) -> bool:
        return monitor not in self.__disabled_monitors

    def enable(self, monitor: MetricMonitor):
        if not self.is_enabled(monitor):
            self.__disabled_monitors.remove(monitor)
            self.on_value_change(self.enabled)

    def disable(self, monitor: MetricMonitor):
        if self.is_enabled(monitor):
            self.__disabled_monitors.add(monitor)
            self.on_value_change(False)

    @abstractmethod
    def on_value_change(self, value: bool):
        pass



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from th2_common.schema.configuration.abstract_configuration import AbstractConfiguration

from typing import List

from th2_common.schema.message.configuration.message_configuration import FieldFilterConfiguration, \
    RouterFilterConfiguration


class GrpcRawRobinStrategy:

    def __init__(self, endpoints, name) -> None:
        self.endpoints = endpoints
        self.name = name


class GrpcServiceConfiguration(AbstractConfiguration):
    pass


class GrpcServerConfiguration(AbstractConfiguration):

    def __init__(self, attributes, host, port, workers, **kwargs) -> None:
        self.attributes = attributes
        self.host = host
        self.port = port
        self.workers = workers
        self.check_unexpected_args(kwargs)


class GrpcEndpointConfiguration(AbstractConfiguration):

    def __init__(self, host, port, attributes, **kwargs) -> None:
        self.host = host
        self.port = port
        self.attributes = attributes
        self.check_unexpected_args(kwargs)


class GrpcConfiguration(AbstractConfiguration):

    def __init__(self, services, server=None, **kwargs) -> None:
        self.services = services
        if server is not None:
            self.serverConfiguration = GrpcServerConfiguration(**server)
        self.check_unexpected_args(kwargs)


class GrpcRouterFilterConfiguration(RouterFilterConfiguration):

    def __init__(self, endpoint: str, metadata, message, **kwargs) -> None:
        self.metadata = metadata
        self.message = message
        self.endpoint = endpoint
        self.check_unexpected_args(kwargs)

    def get_metadata(self) -> List[FieldFilterConfiguration]:
        return self.metadata

    def get_message(self) -> List[FieldFilterConfiguration]:
        return self.message


class GrpcRawFilterStrategy:

    def __init__(self, filters) -> None:
        self.filters = [GrpcRouterFilterConfiguration(**filter_configuration) for filter_configuration in filters]


class GrpcRouterConfiguration(AbstractConfiguration):

    def __init__(self, workers=5, **kwargs):
        self.workers = int(workers)
        self.check_unexpected_args(kwargs)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from google.protobuf.json_format import MessageToJson
from prometheus_client import Counter, Histogram, Gauge
from th2_grpc_common.common_pb2 import MessageGroupBatch

import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_batch_subscriber import AbstractRabbitBatchSubscriber, \
    Metadata
from th2_common.schema.util.util import get_debug_string_group, get_session_alias_and_direction_group, get_sequence
from th2_common.schema.metrics.metric_utils import update_dropped_metrics as util_dropped
from th2_common.schema.metrics.metric_utils import update_total_metrics as util_total


class RabbitMessageGroupBatchSubscriber(AbstractRabbitBatchSubscriber):

    INCOMING_MSG_QUANTITY = Counter('th2_message_subscribe_total',
                                    'Amount of received messages',
                                    common_metrics.DEFAULT_LABELS+(common_metrics.DEFAULT_MESSAGE_TYPE_LABEL_NAME, ))
    INCOMING_MSG_GROUP_QUANTITY = Counter('th2_message_group_subscribe_total',
                                          'Amount of received message groups',
                                          common_metrics.DEFAULT_LABELS)
    INCOMING_MSG_SEQUENCE = Gauge('th2_message_group_sequence_subscribe',
                                  'Last received sequence',
                                  common_metrics.DEFAULT_LABELS)
    INCOMING_MSG_DROPPED_QUANTITY = Counter('th2_message_dropped_subscribe_total',
                                            'Amount of messages dropped after filters',
                                            common_metrics.DEFAULT_LABELS+(common_metrics.DEFAULT_MESSAGE_TYPE_LABEL_NAME, ))
    INCOMING_MSG_GROUP_DROPPED_QUANTITY = Counter('th2_message_group_dropped_subscribe_total',
                                                  'Amount of message groups dropped after filters',
                                                  common_metrics.DEFAULT_LABELS)

    _th2_type = 'MESSAGE_GROUP'

    def update_dropped_metrics(self, batch):
        util_dropped(batch, self.th2_pin, self.INCOMING_MSG_DROPPED_QUANTITY, self.INCOMING_MSG_GROUP_DROPPED_QUANTITY)

    def get_messages(self, batch) -> list:
        return batch.groups

    def extract_metadata(self, message) -> Metadata:
        raise ValueError

    @staticmethod
    def value_from_bytes(body):
        message_group_batch = MessageGroupBatch()
        message_group_batch.ParseFromString(body)
        return message_group_batch

    def to_trace_string(self, value):
        return MessageToJson(value)

    def to_debug_string(self, value):
        return get_debug_string_group(value)

    def update_total_metrics(self, batch):
        util_total(batch, self.th2_pin, self.INCOMING_MSG_QUANTITY, self.INCOMING_MSG_GROUP_QUANTITY, self.INCOMING_MSG_SEQUENCE)



#   Copyright 2021-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC

import th2_common.schema.metrics.metric_monitor as metric_monitor


class Metric(ABC):

    enabled: bool = None

    def create_monitor(self, name: str) -> 'metric_monitor.MetricMonitor':
        """
        Creates new monitor for the metric

        :param name: monitor name
        :return: MetricMonitor which can change metrics value
        """
        pass

    def is_enabled(self, monitor: 'metric_monitor.MetricMonitor') -> bool:
        """
        Checks if status of a monitor is `enabled`

        :param monitor: MetricMonitor
        :return: Status (bool) of MetricMonitor
        """
        pass

    def enable(self, monitor: 'metric_monitor.MetricMonitor'):
        """
        Changes status of a monitor to `enabled`

        :param monitor: MetricMonitor
        """
        pass

    def disable(self, monitor: 'metric_monitor.MetricMonitor'):
        """
        Changes status of a monitor to `disabled`

        :param monitor: MetricMonitor
        """
        pass



#   Copyright 2021-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from th2_common.schema.metrics import common_metrics
from th2_common.schema.util.util import get_session_alias_and_direction_group, get_sequence


def update_total_metrics(batch, th2_pin, message_counter, group_counter, sequence_gauge):
    for group in batch.groups:
        gr_labels = (th2_pin, ) + get_session_alias_and_direction_group(group.messages[0])
        update_message_metrics(group.messages, message_counter, gr_labels)
        if group_counter:
            group_counter.labels(*gr_labels).inc()
        if sequence_gauge:
            sequence_gauge.labels(*gr_labels).set(get_sequence(group))


def update_message_metrics(messages, counter, labels):
    for msg in messages:
        if msg.HasField('raw_message'):
            counter.labels(*labels, common_metrics.TH2_MESSAGE_TYPES['raw']).inc()
        elif msg.HasField('message'):
            counter.labels(*labels, common_metrics.TH2_MESSAGE_TYPES['parsed']).inc()


def update_dropped_metrics(batch, th2_pin, message_counter, group_counter):
    for group in batch.groups:
        labels = (th2_pin, ) + get_session_alias_and_direction_group(group.messages[0])
        update_message_metrics(group.messages, message_counter, labels)
        if group_counter:
            group_counter.labels(*labels).inc()



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
from collections import defaultdict
from threading import Lock
from typing import Callable

from th2_grpc_common.common_pb2 import MessageGroupBatch, Message

from th2_common.schema.exception.router_error import RouterError
from th2_common.schema.filter.strategy.filter_strategy import FilterStrategy
from th2_common.schema.filter.strategy.impl.default_filter_strategy import DefaultFilterStrategy
from th2_common.schema.message.configuration.message_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_listener import MessageListener
from th2_common.schema.message.message_router import MessageRouter
from th2_common.schema.message.message_sender import MessageSender
from th2_common.schema.message.message_subscriber import MessageSubscriber
from th2_common.schema.message.subscriber_monitor import SubscriberMonitor
from th2_common.schema.util.util import get_debug_string_group, get_filters


class SubscriberMonitorImpl(SubscriberMonitor):

    def __init__(self, subscriber: MessageSubscriber, lock=Lock()) -> None:
        self.lock = lock
        self.subscriber = subscriber

    def unsubscribe(self):
        with self.lock:
            self.subscriber.close()


class MultiplySubscribeMonitorImpl(SubscriberMonitor):

    def __init__(self, subscriber_monitors: list) -> None:
        self.subscriber_monitors = subscriber_monitors

    def unsubscribe(self):
        for monitor in self.subscriber_monitors:
            monitor.unsubscribe()


class AbstractRabbitMessageRouter(MessageRouter, ABC):

    def __init__(self, connection_manager, configuration, box_configuration) -> None:
        super().__init__(connection_manager, configuration, box_configuration)
        self._filter_strategy = DefaultFilterStrategy()
        self.queue_connections = list()  # List of queue aliases, which configurations we used to create senders/subs.
        self.queue_connections_lock = Lock()
        self.subscribers = dict()  # queue_alias: subscriber-like objects.
        self.senders = dict()

    @property
    @abstractmethod
    def required_subscribe_attributes(self):
        pass

    @property
    @abstractmethod
    def required_send_attributes(self):
        pass

    def add_subscribe_attributes(self, queue_attr):
        return self.required_subscribe_attributes.union(queue_attr)

    def add_send_attributes(self, queue_attr):
        return self.required_send_attributes.union(queue_attr)

    def _subscribe_by_alias(self, callback: MessageListener, queue_alias) -> SubscriberMonitor:
        subscriber: MessageSubscriber = self.get_subscriber(queue_alias)
        subscriber.add_listener(callback)
        try:
            subscriber.start()
        except Exception as e:
            raise RuntimeError('Can not start subscriber', e)
        return SubscriberMonitorImpl(subscriber, Lock())

    def subscribe(self, callback: MessageListener, *queue_attr) -> SubscriberMonitor:
        attrs = self.add_subscribe_attributes(queue_attr)
        queues = self.configuration.find_queues_by_attr(attrs)
        if len(queues) != 1:
            raise RouterError(
                f'Wrong amount of queues for subscribe. '
                f'Found {len(queues)} queues, but must be only 1. Search was done by {queue_attr} attributes')
        return self._subscribe_by_alias(callback, next(iter(queues.keys())))

    def subscribe_all(self, callback: MessageListener, *queue_attr) -> SubscriberMonitor:
        attrs = self.add_subscribe_attributes(queue_attr)
        subscribers = []
        for queue_alias in self.configuration.find_queues_by_attr(attrs).keys():
            subscribers.append(self._subscribe_by_alias(callback, queue_alias))
        if len(subscribers) == 0:
            raise RouterError(f'Wrong amount of queues for subscribe_all. Must not be empty. '
                              f'Search was done by {attrs} attributes')
        return MultiplySubscribeMonitorImpl(subscribers)

    def unsubscribe_all(self):
        with self.queue_connections_lock:
            for queue in self.queue_connections:
                self.close_connection(queue)
            self.queue_connections.clear()

    def close(self):
        self.unsubscribe_all()

    def send(self, message, *queue_attr):
        attrs = self.add_send_attributes(queue_attr)
        self.filter_and_send(message, attrs, lambda x: None if len(x) == 1 else Exception(f'Found incorrect number of '
                                                                                          f'pins {list(x.keys())} to the send'
                                                                                          f'operation by attributes '
                                                                                          f'{attrs} and filters, '
                                                                                          f'expected 1, actual {len(x)}. '
                                                                                          f'Message: {get_debug_string_group(message)}.'
                                                                                          f'Filters: {get_filters(self.configuration, x.keys())}'))

    def send_all(self, message, *queue_attr):
        attrs = self.add_send_attributes(queue_attr)
        self.filter_and_send(message, attrs, lambda x: None if len(x) != 0 else Exception(f'Found incorrect number of '
                                                                                          f'pins {list(x.keys())} to the send_all '
                                                                                          f'operation by attributes '
                                                                                          f'{attrs} and filters, '
                                                                                          f'expected non-zero, actual {len(x)}. '
                                                                                          f'Message: {get_debug_string_group(message)}.'
                                                                                          f'Filters: {get_filters(self.configuration, x.keys())}'))

    def filter_and_send(self, message, attrs, check: Callable):
        aliases_found_by_attrs = self.configuration.find_queues_by_attr(attrs)
        aliases_to_messages = self.split_and_filter(aliases_found_by_attrs, message)
        result_check = check(aliases_to_messages)
        if result_check is not None:
            raise result_check
        for alias, message in aliases_to_messages.items():
            try:
                sender = self.get_sender(alias)
                sender.start()
                sender.send(message)
            except Exception:
                raise RouterError('Can not start sender')

    def split_and_filter(self, queue_aliases_to_configs, batch) -> dict:
        result = defaultdict(MessageGroupBatch)
        for message in self._get_messages(batch):
            dropped_on_aliases = set()
            aliases_suitable_for_message_part = set()
            for queue_alias, queue_config in queue_aliases_to_configs.items():
                filters = queue_config.filters
                if len(filters) == 0 or self._filter_strategy.verify(message, router_filters=filters):
                    aliases_suitable_for_message_part.add(queue_alias)
                else:
                    dropped_on_aliases.add(queue_alias)
            for queue_alias in aliases_suitable_for_message_part:
                self._add_message(result[queue_alias], message)
            self.update_dropped_metrics(MessageGroupBatch(groups=[message]), dropped_on_aliases)
        return result

    @property
    def filter_strategy(self):
        return self._filter_strategy

    @filter_strategy.setter
    def filter_strategy(self, filter_strategy: FilterStrategy):
        self._filter_strategy = filter_strategy

    def get_subscriber(self, queue_alias) -> MessageSubscriber:
        queue_configuration = self.configuration.get_queue_by_alias(queue_alias)
        with self.queue_connections_lock:
            if queue_alias not in self.queue_connections:
                self.queue_connections.append(queue_alias)
        if not queue_configuration.can_read:
            raise RouterError('Reading from this queue is not allowed')
        with self.subscriber_lock:
            if queue_alias not in self.subscribers or self.subscribers[queue_alias].is_close():
                self.subscribers[queue_alias] = self.create_subscriber(self.connection_manager, queue_configuration,
                                                                       th2_pin=queue_alias)
            return self.subscribers[queue_alias]

    def get_sender(self, queue_alias) -> MessageSender:
        queue_configuration = self.configuration.get_queue_by_alias(queue_alias)
        with self.queue_connections_lock:
            if queue_alias not in self.queue_connections:
                self.queue_connections.append(queue_alias)
        if not queue_configuration.can_write:
            raise RouterError('Writing to this queue is not allowed')
        with self.sender_lock:
            if queue_alias not in self.senders or self.senders[queue_alias].is_close():
                self.senders[queue_alias] = self.create_sender(self.connection_manager, queue_configuration,
                                                               th2_pin=queue_alias)
            return self.senders[queue_alias]

    def close_connection(self, queue_alias):
        with self.subscriber_lock:
            if queue_alias in self.subscribers and not self.subscribers[queue_alias].is_close():
                self.subscribers[queue_alias].close()
        with self.sender_lock:
            if queue_alias in self.senders and not self.senders[queue_alias].is_close():
                self.senders[queue_alias].close()

    @abstractmethod
    def create_sender(self, connection_manager: ConnectionManager,
                      queue_configuration: QueueConfiguration, th2_pin: str) -> MessageSender:
        pass

    @abstractmethod
    def create_subscriber(self, connection_manager: ConnectionManager,
                          queue_configuration: QueueConfiguration, th2_pin: str) -> MessageSubscriber:
        pass

    @abstractmethod
    def _get_messages(self, batch) -> list:
        pass

    @abstractmethod
    def _create_batch(self):
        pass

    @abstractmethod
    def _add_message(self, batch, message):
        pass

    @abstractmethod
    def update_dropped_metrics(self, batch, pin):
        pass



import enum


class QueueAttribute(enum.Enum):
    FIRST = 'first'
    SECOND = 'second'
    SUBSCRIBE = 'subscribe'
    PUBLISH = 'publish'
    PARSED = 'parsed'
    RAW = 'raw'
    EVENT = 'event'
    STORE = 'store'



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from abc import ABC, abstractmethod
from typing import List

from google.protobuf.message import Message

from th2_common.schema.message.configuration.message_configuration import RouterFilterConfiguration


class FilterStrategy(ABC):

    @abstractmethod
    def verify(self, message: Message, router_filter: RouterFilterConfiguration = None,
               router_filters: List[RouterFilterConfiguration] = None):
        pass



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import datetime
import functools
import logging
import threading
import time
from typing import Dict, Optional

import pika
from pika import SelectConnection
from pika.channel import Channel

from th2_common.schema.message.configuration.message_configuration import ConnectionManagerConfiguration
from th2_common.schema.message.impl.rabbitmq.configuration.rabbitmq_configuration import RabbitMQConfiguration

logger = logging.getLogger(__name__)


class Consumer:

    def __init__(self, parameters: pika.ConnectionParameters, prefetch_count,
                 consuming: Dict[str, bool], subscribers: Dict[str, tuple]):
        self.should_reconnect = False
        self.was_consuming = False

        self._parameters = parameters
        self._prefetch_count = prefetch_count
        self._consuming: Dict[str, bool] = consuming
        self._subscribers: Dict[str, tuple] = subscribers

        self._connection: Optional[SelectConnection] = None
        self._channel: Optional[Channel] = None
        self._closing = False

        self._subscribe_allowed = threading.Event()
        self.subscribers_lock = threading.Lock()

    def connect(self):
        logger.info('Connecting by ReconnectingConsumer')
        return pika.SelectConnection(
            parameters=self._parameters,
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed)

    def on_connection_open(self, _unused_connection):
        logger.info('Consumer\'s connection opened')
        self.open_channel()

    def close_connection(self):
        self._subscribe_allowed.clear()
        for consumer_tag in self._consuming.keys():
            self._consuming[consumer_tag] = False
        if self._connection.is_closing or self._connection.is_closed:
            logger.info('Consumer\'s connection is closing or already closed')
        else:
            logger.info('Closing Consumer\'s connection')
            self._connection.close()

    def on_connection_open_error(self, _unused_connection, err):
        self._subscribe_allowed.clear()
        logger.error('Consumer\'s connection open failed: %s', err)
        self.reconnect()

    def on_connection_closed(self, _unused_connection, reason):
        self._subscribe_allowed.clear()
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            logger.warning('Consumer\'s connection closed, reconnect necessary: %s', reason)
            self.reconnect()

    def reconnect(self):
        self.should_reconnect = True
        self.stop()

    def open_channel(self):
        logger.info('Creating the Consumer\'s channel')
        try:
            self._connection.channel(on_open_callback=self.on_channel_open)
        except Exception:
            logger.exception('An error occurred while creating the Consumer\'s channel ')

    def on_channel_open(self, channel):
        logger.info('Consumer\'s channel opened')
        self._channel = channel
        try:
            self._channel.add_on_close_callback(self.on_channel_closed)
        except Exception:
            logger.exception('An error occurred while adding a callback to close the Consumer\'s channel')
        self.set_qos()

    def on_channel_closed(self, channel, reason):
        self._subscribe_allowed.clear()
        logger.warning('Consumer\'s channel %i was closed: %s', channel, reason)
        self.close_connection()

    def set_qos(self):
        try:
            self._channel.basic_qos(prefetch_count=self._prefetch_count,
                                    callback=self.on_basic_qos_ok)
        except Exception:
            logger.exception('An error occurred when specifying prefetch_count for the Consumer\'s channel')

    def on_basic_qos_ok(self, _unused_frame):
        logger.info('QOS set to: %d', self._prefetch_count)
        try:
            self._channel.add_on_cancel_callback(self.on_consumer_cancelled)
            self._subscribe_allowed.set()
            with self.subscribers_lock:
                for consumer_tag in self._subscribers.keys():
                    self.start_consuming(consumer_tag)
        except Exception:
            logger.exception('An error occurred when executing basic_consume by %s for the Consumer\'s channel')

    def start_consuming(self, consumer_tag):
        if not self._subscribe_allowed.is_set():
            logger.warning('Waiting for the Consumer\'s channel to be ready to execute basic_consume')
        self._subscribe_allowed.wait()
        if not self._consuming[consumer_tag]:
            self._channel.basic_consume(queue=self._subscribers[consumer_tag][0],
                                        consumer_tag=consumer_tag,
                                        on_message_callback=self._subscribers[consumer_tag][1])
            self.was_consuming = True
            self._consuming[consumer_tag] = True

    def on_consumer_cancelled(self, method_frame):
        logger.info('Consumer was cancelled remotely, shutting down: %r', method_frame)
        if self._channel:
            self._channel.close()

    def stop_consuming(self, consumer_tag):
        if self._channel:
            logger.info('Sending a Basic.Cancel RPC command to RabbitMQ for tag: %s', consumer_tag)
            cb = functools.partial(self.on_cancel_ok, userdata=consumer_tag)
            self._channel.basic_cancel(consumer_tag, cb)

    def on_cancel_ok(self, _unused_frame, userdata):
        self._consuming[userdata] = False
        logger.info('RabbitMQ acknowledged the cancellation of the consumer: %s', userdata)
        for consumer_tag in self._consuming.keys():
            if self._consuming[consumer_tag] is True:
                return
        self.close_channel()

    def close_channel(self):
        logger.info('Closing the Consumer channel')
        self._channel.close()

    def run(self):
        self._connection = self.connect()
        self._connection.ioloop.start()

    def stop(self):
        if not self._closing:
            self._closing = True
            logger.info('Stopping ReconnectingConsumer')
            for consumer_tag in self._consuming.keys():
                if self._consuming:
                    self.stop_consuming(consumer_tag)
            else:
                self._connection.ioloop.stop()
            logger.info('Stopped ReconnectingConsumer')

    def add_callback_threadsafe(self, cb):
        self._connection.ioloop.add_callback_threadsafe(cb)


class ReconnectingConsumer(object):
    def __init__(self, configuration: RabbitMQConfiguration,
                 connection_manager_configuration: ConnectionManagerConfiguration,
                 connection_parameters: pika.ConnectionParameters):
        self._configuration: RabbitMQConfiguration = configuration
        self._connection_manager_configuration: ConnectionManagerConfiguration = connection_manager_configuration
        self._connection_parameters: pika.ConnectionParameters = connection_parameters
        self._consuming: Dict[str, bool] = dict()
        self._subscribers: Dict[str, tuple] = dict()
        self._reconnect_delay = 0

        self._subscriber_name = self._connection_manager_configuration.subscriber_name

        self._consumer = Consumer(self._connection_parameters, self._connection_manager_configuration.prefetch_count,
                                  self._consuming, self._subscribers)
        self._is_running = True
        self.__next_id_val = -1
        self.__next_id_lock = threading.Lock()

    def next_id(self):
        with self.__next_id_lock:
            self.__next_id_val += 1
            return self.__next_id_val

    def run(self):
        while self._is_running:
            try:
                self._consumer.run()
            finally:
                self._maybe_reconnect()

    def stop(self):
        self._is_running = False
        self._consumer.should_reconnect = False
        self._consumer.stop()

    def add_subscriber(self, queue, on_message_callback):
        with self._consumer.subscribers_lock:
            if self._subscriber_name is None:
                self._subscriber_name = 'rabbit_mq_subscriber'
                logger.info(f"Using default subscriber name: '{self._subscriber_name}'")
            consumer_tag = f'{self._subscriber_name}.{self.next_id()}.{datetime.datetime.now()}'
            self._subscribers[consumer_tag] = (queue, on_message_callback)
            self._consuming[consumer_tag] = False
            self._consumer.start_consuming(consumer_tag)
            return consumer_tag

    def remove_subscriber(self, consumer_tag):
        with self._consumer.subscribers_lock:
            self._consumer.stop_consuming(consumer_tag)
            self._subscribers.pop(consumer_tag)
            self._consuming.pop(consumer_tag)

    def add_callback_threadsafe(self, cb):
        self._consumer.add_callback_threadsafe(cb)

    def _maybe_reconnect(self):
        if self._consumer.should_reconnect:
            self._consumer.stop()
            reconnect_delay = self._get_reconnect_delay()
            logger.info('Reconnecting consumer after %d seconds', reconnect_delay)
            time.sleep(reconnect_delay)
            self._consumer = Consumer(self._connection_parameters, self._connection_manager_configuration.prefetch_count,
                                      self._consuming, self._subscribers)

    def _get_reconnect_delay(self):
        if self._consumer.was_consuming:
            self._reconnect_delay = 0
        else:
            self._reconnect_delay += 1
        if self._reconnect_delay > 30:
            self._reconnect_delay = 30
        return self._reconnect_delay



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from collections import defaultdict

from th2_grpc_common.common_pb2 import EventBatch

from th2_common.schema.event.event_batch_sender import EventBatchSender
from th2_common.schema.event.event_batch_subscriber import EventBatchSubscriber
from th2_common.schema.message.configuration.message_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_message_router import AbstractRabbitMessageRouter
from th2_common.schema.message.impl.rabbitmq.configuration.subscribe_target import SubscribeTarget
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_sender import MessageSender
from th2_common.schema.message.message_subscriber import MessageSubscriber
from th2_common.schema.message.queue_attribute import QueueAttribute


class EventBatchRouter(AbstractRabbitMessageRouter):

    def check_book_name(self, message):
        for event in message.events:
            book = event.id.book_name
            if not book:
                if self.box_configuration.book_name is None:
                    raise Exception('Book name is undefined both explicitly in message IDs and implicitly in box configuration')
                book = self.box_configuration.book_name
                event.id.book_name = book
            if any(message_id.book_name != book for message_id in event.attached_message_ids):
                raise Exception('MessageID attached to event has different book name')
        return message

    def send(self, message, *queue_attr):
        super().send(self.check_book_name(message), *queue_attr)

    def send_all(self, message, *queue_attr):
        super().send_all(self.check_book_name(message), *queue_attr)

    def _get_messages(self, batch) -> list:
        return batch.events

    def _create_batch(self):
        return EventBatch()

    def _add_message(self, batch, message):
        batch.events.append(message)

    def update_dropped_metrics(self, batch, pin):
        pass

    def split_and_filter(self, queue_aliases_to_configs, batch) -> dict:
        result = defaultdict(EventBatch)
        for message in self._get_messages(batch):
            for queue_alias in queue_aliases_to_configs.keys():
                self._add_message(result[queue_alias], message)
        return result

    @property
    def required_subscribe_attributes(self):
        return {QueueAttribute.SUBSCRIBE.value, QueueAttribute.EVENT.value}

    @property
    def required_send_attributes(self):
        return {QueueAttribute.PUBLISH.value, QueueAttribute.EVENT.value}

    def _find_by_filter(self, queues: {str: QueueConfiguration}, msg) -> dict:
        return {key: msg for key in queues.keys()}

    def create_sender(self, connection_manager: ConnectionManager,
                      queue_configuration: QueueConfiguration, th2_pin) -> MessageSender:
        return EventBatchSender(connection_manager, queue_configuration.exchange, queue_configuration.routing_key,
                                th2_pin=th2_pin)

    def create_subscriber(self, connection_manager: ConnectionManager,
                          queue_configuration: QueueConfiguration, th2_pin) -> MessageSubscriber:
        subscribe_target = SubscribeTarget(queue_configuration.queue, queue_configuration.routing_key)
        return EventBatchSubscriber(connection_manager, queue_configuration, subscribe_target, th2_pin=th2_pin)



#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import threading

import pika
from google.protobuf.pyext._message import SetAllowOversizeProtos

from th2_common.schema.message.configuration.message_configuration import ConnectionManagerConfiguration
from th2_common.schema.message.impl.rabbitmq.configuration.rabbitmq_configuration import RabbitMQConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.reconnecting_consumer import ReconnectingConsumer
from th2_common.schema.message.impl.rabbitmq.connection.reconnecting_publisher import ReconnectingPublisher
from th2_common.schema.metrics.common_metrics import HealthMetrics


logger = logging.getLogger(__name__)


class ConnectionManager:

    def __init__(self, configuration: RabbitMQConfiguration,
                 connection_manager_configuration: ConnectionManagerConfiguration) -> None:
        self.__credentials = pika.PlainCredentials(configuration.username,
                                                   configuration.password)
        self.__connection_parameters = pika.ConnectionParameters(virtual_host=configuration.vhost,
                                                                 host=configuration.host,
                                                                 port=configuration.port,
                                                                 credentials=self.__credentials)

        SetAllowOversizeProtos(connection_manager_configuration.message_recursion_limit > 100)

        self.__metrics = HealthMetrics(self)

        self.consumer = ReconnectingConsumer(configuration,
                                             connection_manager_configuration,
                                             self.__connection_parameters)
        threading.Thread(target=self.consumer.run).start()

        self.publisher = ReconnectingPublisher(self.__connection_parameters)
        threading.Thread(target=self.publisher.run).start()

        self.__metrics.enable()

    def close(self):
        try:
            self.consumer.stop()
        except Exception:
            logger.exception("Error while stopping Consumer")
        try:
            self.publisher.stop()
        except Exception:
            logger.exception("Error while stopping Publisher")
        self.__metrics.disable()



#   Copyright 2021-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import th2_common.schema.metrics.metric as metric


class MetricMonitor:

    def __init__(self, name: str, arbiter: 'metric.Metric') -> None:
        self.name = name
        self.arbiter = arbiter

    @property
    def is_enabled(self) -> bool:
        return self.arbiter.is_enabled(self)

    @is_enabled.setter
    def is_enabled(self, value):
        if value:
            self.enable()
        else:
            self.disable()

    @property
    def is_metric_enabled(self):
        return self.arbiter.enabled

    def enable(self):
        self.arbiter.enable(self)

    def disable(self):
        self.arbiter.disable(self)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from th2_grpc_common.common_pb2 import AnyMessage, MessageGroup, MessageGroupBatch, RawMessageBatch

from th2_common.schema.message.impl.rabbitmq.group.rabbit_message_group_batch_router_adapter import \
    RabbitMessageGroupBatchRouterAdapter
from th2_common.schema.message.queue_attribute import QueueAttribute


class RabbitRawBatchRouter(RabbitMessageGroupBatchRouterAdapter):

    @property
    def required_subscribe_attributes(self):
        return {QueueAttribute.SUBSCRIBE.value, QueueAttribute.RAW.value}

    @property
    def required_send_attributes(self):
        return {QueueAttribute.PUBLISH.value, QueueAttribute.RAW.value}

    @staticmethod
    def to_group_batch(message):
        messages = [AnyMessage(raw_message=msg) for msg in message.messages]
        group = MessageGroup(messages=messages)
        value = MessageGroupBatch(groups=[group])
        return value

    @staticmethod
    def from_group_batch(message):
        return RawMessageBatch(messages=[anymsg.raw_message for group in message.groups for anymsg in group.messages if
                                         anymsg.HasField('raw_message')])
